#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Multiple entities entity and resolver."""

from __future__ import annotations

import dataclasses
import json
from typing import TYPE_CHECKING, Any

from invenio_records_resources.references.entity_resolvers import EntityProxy
from invenio_records_resources.references.entity_resolvers.base import EntityResolver
from invenio_requests.resolvers.registry import ResolverRegistry

if TYPE_CHECKING:
    from flask_principal import Identity, Need


@dataclasses.dataclass
class MultipleEntitiesEntity:
    """Entity representing multiple entities."""

    entities: list[EntityProxy]


class MultipleEntitiesProxy(EntityProxy):
    """Proxy for multiple-entities entity."""

    def _resolve(self) -> MultipleEntitiesEntity:
        """Resolve the entity reference into entity."""
        values = json.loads(self._parse_ref_dict_id())
        return MultipleEntitiesEntity(
            entities=[
                ResolverRegistry.resolve_entity_proxy(ref, raise_=True)  # type: ignore
                for ref in values
            ]
        )

    def get_needs(self, ctx: dict | None = None) -> list[Need]:
        """Get needs that the entity generate."""
        ret = []
        for entity in self._resolve().entities:
            ret.extend(entity.get_needs(ctx) or [])
        return ret

    def pick_resolved_fields(self, identity: Identity, resolved_dict: dict) -> dict:
        """Pick resolved fields for serialization of the entity to json."""
        return {"multiple": resolved_dict["id"]}


class MultipleEntitiesResolver(EntityResolver):
    """A resolver that resolves multiple entities entity."""

    type_id = "multiple"

    def __init__(self) -> None:
        """Initialize the resolver."""
        super().__init__("multiple")

    def matches_reference_dict(self, ref_dict: dict) -> bool:
        """Check if the reference dictionary can be resolved by this resolver."""
        return self._parse_ref_dict_type(ref_dict) == self.type_id

    def _reference_entity(self, entity: MultipleEntitiesEntity) -> dict[str, str]:
        """Return a reference dictionary for the entity."""
        print("!!!! multiple_entities/__init__.py _reference_entity", entity.entities)
        return {
            self.type_id: json.dumps([part.reference_dict for part in entity.entities])
        }

    def matches_entity(self, entity: Any) -> bool:
        """Check if the entity can be serialized to a reference by this resolver."""
        return isinstance(entity, MultipleEntitiesEntity)

    def _get_entity_proxy(self, ref_dict: dict) -> MultipleEntitiesProxy:
        """Get the entity proxy for the reference dictionary.

        Note: the proxy is returned to ensure the entity is loaded lazily, when needed.
        :param ref_dict: Reference dictionary.
        """
        return MultipleEntitiesProxy(self, ref_dict)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Events for workflow requests."""

from __future__ import annotations

import dataclasses
from functools import cached_property
from typing import TYPE_CHECKING

from oarepo_workflows.requests.generators.multiple_entities import (
    MultipleEntitiesGenerator,
)

if TYPE_CHECKING:
    from invenio_records_permissions.generators import Generator


@dataclasses.dataclass
class WorkflowEvent:
    """Class representing a workflow event."""

    submitters: list[Generator] | tuple[Generator]
    """List of submitters to be used for the event.

       The generators supply needs. The user must have at least one of the needs 
       to be able to create a workflow event.
    """

    @cached_property
    def submitter_generator(self) -> Generator:
        """Return the requesters as a single requester generator."""
        return MultipleEntitiesGenerator(self.submitters)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Need generators."""

from __future__ import annotations

from .auto import AutoApprove, AutoRequest, auto_approve_need, auto_request_need
from .conditionals import IfEventType, IfRequestType, IfRequestTypeBase
from .multiple_entities import MultipleEntitiesGenerator
from .recipient_generator import RecipientGeneratorMixin

__all__ = (
    "AutoApprove",
    "auto_approve_need",
    "AutoRequest",
    "auto_request_need",
    "IfEventType",
    "IfRequestType",
    "IfRequestTypeBase",
    "MultipleEntitiesGenerator",
    "RecipientGeneratorMixin",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Permissions for workflows."""

from __future__ import annotations

from .generators import FromRecordWorkflow, IfInState, WorkflowPermission
from .record_permission_policy import WorkflowRecordPermissionPolicy
from .workflow_permissions import DefaultWorkflowPermissions

__all__ = (
    "IfInState",
    "WorkflowPermission",
    "WorkflowRecordPermissionPolicy",
    "DefaultWorkflowPermissions",
    "FromRecordWorkflow",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Mixin for permission generators that can be used as recipients in WorkflowRequest."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from invenio_records_resources.records import Record
    from invenio_requests.customizations import RequestType


class RecipientGeneratorMixin:
    """Mixin for permission generators that can be used as recipients in WorkflowRequest."""

    def reference_receivers(
        self,
        record: Optional[Record] = None,
        request_type: Optional[RequestType] = None,
        **context: Any,
    ) -> list[dict[str, str]]:  # pragma: no cover
        """Return the reference receiver(s) of the request.

        This call requires the context to contain at least "record" and "request_type"

        Must return a list of dictionary serialization of the receivers.

        Might return empty list or None to indicate that the generator does not
        provide any receivers.
        """
        raise NotImplementedError("Implement reference receiver in your code")



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Workflow system field."""

from __future__ import annotations

from typing import Protocol

from invenio_records.systemfields.model import ModelField
from oarepo_runtime.records.systemfields import MappingSystemFieldMixin


class WithWorkflow(Protocol):
    """A protocol for a record's parent containing a workflow field.

    Later on, if typing.Intersection is implemented,
    one could use it to have the record correctly typed as
    record: Intersection[WithWorkflow, ParentRecord]
    """

    workflow: str
    """Workflow of the record."""


class WorkflowField(MappingSystemFieldMixin, ModelField):
    """Workflow system field, should be defined on ParentRecord."""

    def __init__(self) -> None:
        """Initialize the workflow field."""
        self._workflow = None  # added in db
        super().__init__(model_field_name="workflow", key="workflow")

    @property
    def mapping(self) -> dict[str, dict[str, str]]:
        """Elasticsearch mapping."""
        return {self.attr_name: {"type": "keyword"}}



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""API blueprints."""

from __future__ import annotations

from typing import TYPE_CHECKING

from flask import Blueprint, Flask

if TYPE_CHECKING:
    from flask.blueprints import BlueprintSetupState


def create_api_blueprint(app: Flask) -> Blueprint:
    """Create requests blueprint."""
    blueprint = Blueprint("oarepo-workflows", __name__)

    # noinspection PyUnusedLocal
    def register_auto_approve_entity_resolver(state: BlueprintSetupState) -> None:
        from oarepo_workflows.resolvers.auto_approve import AutoApproveResolver

        requests = app.extensions["invenio-requests"]
        requests.entity_resolvers_registry.register_type(AutoApproveResolver())

    blueprint.record_once(register_auto_approve_entity_resolver)

    return blueprint



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Support any workflows on Invenio record."""

from __future__ import annotations

from oarepo_workflows.services.permissions import (
    FromRecordWorkflow,
    IfInState,
    WorkflowPermission,
    WorkflowRecordPermissionPolicy,
)

from .base import Workflow
from .requests import (
    AutoApprove,
    AutoRequest,
    WorkflowRequest,
    WorkflowRequestEscalation,
    WorkflowRequestPolicy,
    WorkflowTransitions,
)

__all__ = (
    "IfInState",
    "Workflow",
    "WorkflowPermission",
    "WorkflowRecordPermissionPolicy",
    "WorkflowRequestPolicy",
    "WorkflowRequest",
    "WorkflowRequestEscalation",
    "WorkflowTransitions",
    "FromRecordWorkflow",
    "AutoApprove",
    "AutoRequest",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Definition of workflow permissions."""

from __future__ import annotations

from typing import Any

from invenio_records_permissions import RecordPermissionPolicy
from invenio_records_permissions.generators import (
    AuthenticatedUser,
    Disable,
    SystemProcess,
)
from oarepo_runtime.services.permissions import RecordOwners

from .generators import IfInState, SameAs


class DefaultWorkflowPermissions(RecordPermissionPolicy):
    """Base class for workflow permissions, subclass from it and put the result to Workflow constructor.

    Example:
        class MyWorkflowPermissions(DefaultWorkflowPermissions):
            can_read = [AnyUser()]
    in invenio.cfg
    WORKFLOWS = {
        'default': Workflow(
            permission_policy_cls = MyWorkflowPermissions, ...
        )
    }

    """

    # TODO: new version - update; edit current version - disable -> idk if there's other way than something like IfNoEditDraft/IfNoNewVersionDraft generators-

    files_edit = [
        IfInState("draft", [RecordOwners()]),
        IfInState("published", [Disable()]),
    ]

    system_process = SystemProcess()

    def __init__(self, action_name: str | None = None, **over: Any) -> None:
        """Initialize the workflow permissions."""
        can = getattr(self, f"can_{action_name}")
        if self.system_process not in can:
            can.append(self.system_process)
        over["policy"] = self
        super().__init__(action_name, **over)

    can_read = [
        IfInState("draft", [RecordOwners()]),
        IfInState("published", [AuthenticatedUser()]),
    ]
    # can_read_deleted is used by RDM. As long as workflows are
    # not intended for rdm records only, we need to keep this permission
    # simple and the implementation must use RDM-based implementation
    #
    # TODO: we should provide somewhere RDM aware workflow permissions
    # so that repositories can use them out of the box
    can_read_deleted = [SystemProcess()]
    can_update = [IfInState("draft", [RecordOwners()])]
    can_delete = [
        IfInState("draft", [RecordOwners()]),
    ]
    can_create = [AuthenticatedUser()]
    can_publish = [AuthenticatedUser()]
    can_new_version = [AuthenticatedUser()]

    can_create_files = [SameAs("files_edit")]
    can_set_content_files = [SameAs("files_edit")]
    can_commit_files = [SameAs("files_edit")]
    can_update_files = [SameAs("files_edit")]
    can_delete_files = [SameAs("files_edit")]
    can_read_files = [SameAs("can_read")]
    can_get_content_files = [SameAs("can_read")]
    can_list_files = [SameAs("can_read")]
    can_manage_files = [Disable()]

    can_read_draft = [SameAs("can_read")]
    can_update_draft = [SameAs("can_update")]
    can_delete_draft = [SameAs("can_delete")]

    # RDM records permissions
    can_bulk_add = [Disable()]

    # draft files
    can_draft_commit_files = [SameAs("can_update_draft")]
    can_draft_delete_files = [SameAs("can_update_draft")]
    can_draft_get_content_files = [SameAs("can_read_draft")]
    can_draft_read_files = [SameAs("can_read_draft")]
    can_draft_set_content_files = [SameAs("can_update_draft")]
    can_draft_update_files = [SameAs("can_update_draft")]
    can_draft_create_files = [SameAs("files_edit")]

    # draft media files
    can_draft_media_commit_files = [SameAs("can_update_draft")]
    can_draft_media_create_files = [SameAs("can_update_draft")]
    can_draft_media_delete_files = [SameAs("can_update_draft")]
    can_draft_media_get_content_files = [SameAs("can_read_draft")]
    can_draft_media_read_files = [SameAs("can_read_draft")]
    can_draft_media_set_content_files = [SameAs("can_update_draft")]
    can_draft_media_update_files = [SameAs("can_update_draft")]

    # published media files
    can_media_commit_files = [Disable()]
    can_media_create_files = [Disable()]
    can_media_delete_files = [Disable()]
    can_media_get_content_files = [SameAs("can_read_files")]
    can_media_read_deleted_files = [Disable()]
    can_media_read_files = [SameAs("can_read_files")]
    can_media_set_content_files = [Disable()]
    can_media_update_files = [Disable()]

    # record
    can_lift_embargo = [SameAs("can_manage")]

    # record management
    can_manage = [RecordOwners()]

    can_manage_internal = [SystemProcess()]

    can_manage_quota = [SystemProcess()]
    can_manage_record_access = [SystemProcess()]

    can_moderate = [Disable()]

    can_preview = [SameAs("can_read")]
    can_review = [Disable()]
    can_view = [SameAs("can_read")]
    can_purge = [SystemProcess()]

    can_pid_create = [SameAs("can_update")]
    can_pid_delete = [SameAs("can_update")]
    can_pid_discard = [SameAs("can_update")]
    can_pid_manage = [SameAs("can_update")]
    can_pid_register = [SameAs("can_update")]
    can_pid_update = [SameAs("can_update")]

    # stats
    can_query_stats = [Disable()]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Auto approve entity and resolver."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from invenio_records_resources.references.entity_resolvers import EntityProxy
from invenio_records_resources.references.entity_resolvers.base import EntityResolver

if TYPE_CHECKING:
    from flask_principal import Identity, Need


class AutoApproveEntity:
    """Entity representing auto approve."""


class AutoApproveProxy(EntityProxy):
    """Proxy for auto approve entity."""

    def _resolve(self) -> AutoApproveEntity:
        """Resolve the entity reference into entity."""
        return AutoApproveEntity()

    def get_needs(self, ctx: dict | None = None) -> list[Need]:
        """Get needs that the entity generate."""
        return []  # grant_tokens calls this

    def pick_resolved_fields(self, identity: Identity, resolved_dict: dict) -> dict:
        """Pick resolved fields for serialization of the entity to json."""
        return {"auto_approve": resolved_dict["id"]}


class AutoApproveResolver(EntityResolver):
    """A resolver that resolves auto approve entity."""

    type_id = "auto_approve"

    def __init__(self) -> None:
        """Initialize the resolver."""
        super().__init__("auto_approve")

    def matches_reference_dict(self, ref_dict: dict) -> bool:
        """Check if the reference dictionary can be resolved by this resolver."""
        return self._parse_ref_dict_type(ref_dict) == self.type_id

    def _reference_entity(self, entity: Any) -> dict[str, str]:
        """Return a reference dictionary for the entity."""
        return {self.type_id: "true"}

    def matches_entity(self, entity: Any) -> bool:
        """Check if the entity can be serialized to a reference by this resolver."""
        return isinstance(entity, AutoApproveEntity)

    def _get_entity_proxy(self, ref_dict: dict) -> AutoApproveProxy:
        """Get the entity proxy for the reference dictionary.

        Note: the proxy is returned to ensure the entity is loaded lazily, when needed.
        :param ref_dict: Reference dictionary.
        """
        return AutoApproveProxy(self, ref_dict)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Service for reading auto-approve entities.

The implementation is simple as auto approve is just one entity
so there is no need to store it to database/fetch it from the database.
"""

from __future__ import annotations

from oarepo_runtime.services.entity.config import KeywordEntityServiceConfig
from oarepo_runtime.services.entity.service import KeywordEntityService


class AutoApproveEntityServiceConfig(KeywordEntityServiceConfig):
    """Configuration for auto-approve entity service."""

    service_id = "auto_approve"
    keyword = "auto_approve"


class AutoApproveEntityService(KeywordEntityService):
    """Service for reading auto-approve entities."""

    pass



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Proxies for accessing the current OARepo workflows extension without bringing dependencies."""

from __future__ import annotations

from typing import TYPE_CHECKING

from flask import current_app
from werkzeug.local import LocalProxy

if TYPE_CHECKING:
    from oarepo_workflows.ext import OARepoWorkflows

current_oarepo_workflows: OARepoWorkflows = LocalProxy(  # type: ignore
    lambda: current_app.extensions["oarepo-workflows"]
)
"""Proxy to access the current OARepo workflows extension."""



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Generators for needs/excludes/queries based on workflows."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from opensearch_dsl.query import Query

from oarepo_workflows import FromRecordWorkflow, Workflow, WorkflowRequest
from oarepo_workflows.errors import (
    EventTypeNotInWorkflow,
    InvalidWorkflowError,
    MissingWorkflowError,
    RequestTypeNotInWorkflow,
)
from oarepo_workflows.proxies import current_oarepo_workflows

if TYPE_CHECKING:
    from flask_principal import Need
    from invenio_records_permissions.generators import Generator
    from invenio_records_resources.records import Record
    from invenio_requests.customizations import EventType, RequestType
    from invenio_requests.records.api import Request


class MissingTopicError(ValueError):
    """Raised when the topic is missing in the request generator arguments."""

    pass


def _get_workflow_code_from_workflow(workflow: Workflow) -> str:
    """Return the workflow code from the workflow."""
    workflow_code = next(
        iter(
            k
            for k, v in current_oarepo_workflows.record_workflows.items()
            if v is workflow
        )
    )
    return workflow_code


class RequestPolicyWorkflowCreators(FromRecordWorkflow):
    """Base class that generates creators from a workflow request."""

    def _kwargs_parser(self, **kwargs: Any) -> dict[str, Any]:
        """Transform the kwargs for subsequent methods."""
        return kwargs

    def _requester_generator(self, **kwargs: Any) -> Generator:
        """Return the requesters as a single requester generator."""
        raise NotImplementedError()

    def _get_workflow_request(
        self, request_type_id: str, **kwargs: Any
    ) -> tuple[Workflow, WorkflowRequest]:
        """Return the workflow request from the context."""
        if "record" not in kwargs:
            raise MissingTopicError(
                "Topic not found in request permissions generator arguments, can't get workflow."
            )
        record = kwargs["record"]
        workflow = current_oarepo_workflows.get_workflow(record)
        workflow_requests = workflow.requests()
        try:
            workflow_request = workflow_requests[request_type_id]
        except KeyError as e:
            workflow_code = _get_workflow_code_from_workflow(workflow)
            raise RequestTypeNotInWorkflow(
                request_type=request_type_id,
                workflow_code=workflow_code,
            ) from e
        return workflow, workflow_request

    # return empty needs on MissingTopicError
    # match None in query filter
    # excludes empty needs
    def needs(self, **context: Any) -> list[Need]:  # type: ignore
        """Return the needs generated by the workflow permission."""
        try:
            context = self._kwargs_parser(**context)
            generator = self._requester_generator(**context)
            return generator.needs(**context)
        except (MissingWorkflowError, InvalidWorkflowError, MissingTopicError):
            return []

    def excludes(self, **context: Any) -> list[Need]:
        """Return the needs excluded by the workflow permission."""
        try:
            context = self._kwargs_parser(**context)
            generator = self._requester_generator(**context)
            return generator.excludes(**context)
        except (MissingWorkflowError, InvalidWorkflowError, MissingTopicError):
            return []

    # not tested
    def query_filter(
        self, record: Record = None, request_type: RequestType = None, **context: Any
    ) -> Query:
        """Return the query filter generated by the workflow permission."""
        try:
            context = self._kwargs_parser(**context)
            generator = self._requester_generator(
                record=record, request_type=request_type, **context
            )
            return generator.query_filter(
                record=record, request_type=request_type, **context
            )
        except (MissingWorkflowError, InvalidWorkflowError, MissingTopicError):
            return Query("match_none")


class RequestCreatorsFromWorkflow(RequestPolicyWorkflowCreators):
    """Generates creators from a workflow request to be used in the request 'create' permission."""

    def __init__(self) -> None:
        """Initialize the generator."""
        super().__init__(action="create")

    @override
    def _requester_generator(self, **kwargs: Any) -> Generator:
        request_type: RequestType = kwargs["request_type"]
        workflow, workflow_request = self._get_workflow_request(
            request_type.type_id, **kwargs
        )
        return workflow_request.requester_generator


class EventCreatorsFromWorkflow(RequestPolicyWorkflowCreators):
    """Generates creators from a workflow request to be used in the event 'create' permission."""

    def __init__(self) -> None:
        """Initialize the generator."""
        super().__init__(action="create")

    @override
    def _kwargs_parser(self, **kwargs: Any) -> dict[str, Any]:
        request: Request = kwargs["request"]
        kwargs.setdefault("request_type", request.type)
        try:
            kwargs["record"] = request.topic.resolve()
        except Exception as e:  # noqa              TODO: better exception catching here
            raise MissingTopicError(
                "Topic not found in request event permissions generator arguments, can't get workflow."
            ) from e
        return kwargs

    @override
    def _requester_generator(self, **kwargs: Any) -> Generator:
        request_type: RequestType = kwargs["request_type"]
        event_type: EventType = kwargs["event_type"]
        workflow, workflow_request = self._get_workflow_request(
            request_type.type_id,
            **kwargs,
        )
        try:
            workflow_event = workflow_request.allowed_events[event_type.type_id]
        except KeyError as e:
            workflow_code = _get_workflow_code_from_workflow(workflow)
            raise EventTypeNotInWorkflow(
                request_type=request_type.type_id,
                workflow_code=workflow_code,
                event_type=event_type.type_id,
            ) from e
        return workflow_event.submitter_generator



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Conditional generators based on request and event types."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any

from invenio_records_permissions.generators import ConditionalGenerator, Generator

if TYPE_CHECKING:
    from invenio_requests.customizations import EventType, RequestType


class IfRequestTypeBase(abc.ABC, ConditionalGenerator):
    """Base class for conditional generators that generate needs based on request type."""

    def __init__(
        self, request_types: list[str] | tuple[str] | str, then_: list[Generator]
    ) -> None:
        """Initialize the generator."""
        super().__init__(then_, else_=[])
        if not isinstance(request_types, (list, tuple)):
            request_types = [request_types]
        self.request_types = request_types


class IfRequestType(IfRequestTypeBase):
    """Conditional generator that generates needs when a current request is of a given type."""

    def _condition(self, request_type: RequestType, **kwargs: Any) -> bool:
        return request_type.type_id in self.request_types


class IfEventType(ConditionalGenerator):
    """Conditional generator that generates needs when a current event is of a given type."""

    def __init__(
        self,
        event_types: list[str] | tuple[str] | str,
        then_: list[Generator],
        else_: list[Generator] | None = None,
    ) -> None:
        """Initialize the generator."""
        else_ = [] if else_ is None else else_
        super().__init__(then_, else_=else_)
        if not isinstance(event_types, (list, tuple)):
            event_types = [event_types]
        self.event_types = event_types

    def _condition(self, event_type: EventType, **kwargs: Any) -> bool:
        return event_type.type_id in self.event_types



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Requests layer."""

from __future__ import annotations

from .generators import AutoApprove, AutoRequest, RecipientGeneratorMixin
from .policy import (
    WorkflowRequestPolicy,
)
from .requests import (
    RecipientEntityReference,
    WorkflowRequest,
    WorkflowRequestEscalation,
    WorkflowTransitions,
)

__all__ = (
    "WorkflowRequestPolicy",
    "RecipientEntityReference",
    "WorkflowRequest",
    "WorkflowRequestEscalation",
    "WorkflowTransitions",
    "AutoApprove",
    "AutoRequest",
    "RecipientGeneratorMixin",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Errors raised by oarepo-workflows."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from marshmallow import ValidationError

if TYPE_CHECKING:
    from invenio_records_resources.records import Record


def _get_id_from_record(record: Record | dict) -> str:
    """Get the id from a record.

    :param record: A record or a dict representing a record.
    :return str: The id of the record.
    """
    # community record doesn't have id in dict form, only uuid
    try:
        if "id" in record:
            return str(record["id"])
    except TypeError:
        pass
    if hasattr(record, "id"):
        return str(record.id)
    return str(record)


def _format_record(record: Record | dict) -> str:
    """Format a record for error messages.

    :param record: A record or a dict representing a record.
    :return str: A formatted string representing the record.
    """
    return f"{type(record).__name__}[{_get_id_from_record(record)}]"


class MissingWorkflowError(ValidationError):
    """Exception raised when a required workflow is missing."""

    def __init__(self, message: str, record: Record | dict | None = None) -> None:
        """Initialize the exception."""
        self.record = record
        if record:
            super().__init__(f"{message} Used on record {_format_record(record)}")
        else:
            super().__init__(message)


class InvalidWorkflowError(ValidationError):
    """Exception raised when a workflow is invalid."""

    def __init__(
        self,
        message: str,
        record: Record | dict | None = None,
        community_id: Optional[str] = None,
    ) -> None:
        """Initialize the exception."""
        self.record = record
        if record:
            super().__init__(f"{message} Used on record {_format_record(record)}")
        elif community_id:
            super().__init__(f"{message} Used on community {community_id}")
        else:
            super().__init__(message)


class InvalidConfigurationError(Exception):
    """Exception raised when a configuration is invalid."""


class EventTypeNotInWorkflow(Exception):
    """Exception raised when user tries to create a request with a request type that is not defined in the workflow."""

    def __init__(self, request_type: str, event_type: str, workflow_code: str) -> None:
        """Initialize the exception."""
        self.request_type = request_type
        self.workflow = workflow_code
        self.event_type = event_type

    @property
    def description(self) -> str:
        """Exception's description."""
        return f"Event type {self.event_type} is not on request type {self.request_type} in workflow {self.workflow}."


class RequestTypeNotInWorkflow(Exception):
    """Exception raised when user tries to create a request with a request type that is not defined in the workflow."""

    def __init__(self, request_type: str, workflow_code: str) -> None:
        """Initialize the exception."""
        self.request_type = request_type
        self.workflow = workflow_code

    @property
    def description(self) -> str:
        """Exception's description."""
        return f"Request type {self.request_type} not in workflow {self.workflow}."



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Mixin for records with workflow support."""

from __future__ import annotations

import marshmallow as ma
from invenio_drafts_resources.services.records.schema import ParentSchema
from invenio_rdm_records.services.schemas.parent import RDMParentSchema

class WorkflowParentSchema(ParentSchema):
    """Schema for parent record with workflow support."""

    workflow = ma.fields.String()

class RDMWorkflowParentSchema(RDMParentSchema):
    """RDM based schema for parent record with workflow support."""

    workflow = ma.fields.String()



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Auto request and auto approve generators."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from invenio_access import SystemRoleNeed
from invenio_records_permissions.generators import Generator

from .recipient_generator import RecipientGeneratorMixin

if TYPE_CHECKING:
    from flask_principal import Need
    from invenio_records_resources.records import Record
    from invenio_requests.customizations import RequestType

auto_request_need = SystemRoleNeed("auto_request")
auto_approve_need = SystemRoleNeed("auto_approve")


class AutoRequest(Generator):
    """Auto request generator.

    This generator is used to automatically create a request
    when a record is moved to a specific state.
    """

    def needs(self, **context: Any) -> list[Need]:
        """Get needs that signal workflow to automatically create the request."""
        return [auto_request_need]


class AutoApprove(RecipientGeneratorMixin, Generator):
    """Auto approve generator.

    If the generator is used within recipients of a request,
    the request will be automatically approved when the request is submitted.
    """

    def reference_receivers(
        self,
        record: Optional[Record] = None,
        request_type: Optional[RequestType] = None,
        **kwargs: Any,
    ) -> list[dict[str, str]]:
        """Return the reference receiver(s) of the auto-approve request.

        Returning "auto_approve" is a signal to the workflow that the request should be auto-approved.
        """
        return [{"auto_approve": "true"}]

    def needs(self, **context: Any) -> list[Need]:
        """Get needs that signal workflow to automatically approve the request."""
        raise ValueError(
            "Auto-approve generator can not create needs and "
            "should be used only in `recipient` section of WorkflowRequest."
        )

    def excludes(self, **context: Any) -> list[Need]:
        """Get needs that signal workflow to automatically approve the request."""
        raise ValueError(
            "Auto-approve generator can not create needs and "
            "should be used only in `recipient` section of WorkflowRequest."
        )

    def query_filter(self, **context: Any) -> list[dict]:
        """Get needs that signal workflow to automatically approve the request."""
        raise ValueError(
            "Auto-approve generator can not create needs and "
            "should be used only in `recipient` section of WorkflowRequest."
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Flask extension for workflows."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional

import importlib_metadata
from invenio_drafts_resources.services.records.uow import ParentRecordCommitOp
from invenio_records_resources.services.uow import RecordCommitOp, unit_of_work
from oarepo_runtime.datastreams.utils import get_record_service_for_record

from oarepo_workflows.errors import InvalidWorkflowError, MissingWorkflowError
from oarepo_workflows.proxies import current_oarepo_workflows
from oarepo_workflows.services.auto_approve import (
    AutoApproveEntityService,
    AutoApproveEntityServiceConfig,
)
from oarepo_workflows.services.multiple_entities import (
    MultipleEntitiesEntityService,
    MultipleEntitiesEntityServiceConfig,
)

if TYPE_CHECKING:
    from flask import Flask
    from flask_principal import Identity
    from invenio_drafts_resources.records import ParentRecord, Record
    from invenio_records_resources.services.uow import UnitOfWork

    from oarepo_workflows.base import (
        StateChangedNotifier,
        Workflow,
        WorkflowChangeNotifier,
    )
    from oarepo_workflows.records.systemfields.workflow import WithWorkflow


class OARepoWorkflows:
    """OARepo workflows extension."""

    def __init__(self, app: Optional[Flask] = None) -> None:
        """Initialize the extension.

        :param app: Flask application to initialize with.
        If not passed here, it can be passed later using init_app method.
        """
        if app:
            self.init_config(app)
            self.init_app(app)
            self.init_services()

    # noinspection PyMethodMayBeStatic
    def init_config(self, app: Flask) -> None:
        """Initialize configuration.

        :param app: Flask application to initialize with.
        """
        from . import ext_config

        if "OAREPO_PERMISSIONS_PRESETS" not in app.config:
            app.config["OAREPO_PERMISSIONS_PRESETS"] = {}

        for k in ext_config.OAREPO_PERMISSIONS_PRESETS:
            if k not in app.config["OAREPO_PERMISSIONS_PRESETS"]:
                app.config["OAREPO_PERMISSIONS_PRESETS"][k] = (
                    ext_config.OAREPO_PERMISSIONS_PRESETS[k]
                )

        app.config.setdefault("WORKFLOWS", ext_config.WORKFLOWS)

        app.config.setdefault(
            "OAREPO_WORKFLOWS_SET_REQUEST_PERMISSIONS",
            ext_config.OAREPO_WORKFLOWS_SET_REQUEST_PERMISSIONS,
        )

        app.config.setdefault("REQUESTS_ALLOWED_RECEIVERS", []).extend(
            ext_config.WORKFLOWS_ALLOWED_REQUEST_RECEIVERS
        )

    def init_services(self) -> None:
        """Initialize workflow services."""
        # noinspection PyAttributeOutsideInit
        self.auto_approve_service = AutoApproveEntityService(
            config=AutoApproveEntityServiceConfig()
        )
        self.multiple_recipients_service = MultipleEntitiesEntityService(
            config=MultipleEntitiesEntityServiceConfig()
        )

    @cached_property
    def state_changed_notifiers(self) -> list[StateChangedNotifier]:
        """Return a list of state changed notifiers.

        State changed notifiers are callables that are called when a state of a record changes,
        for example as a result of a workflow transition.

        They are registered as entry points in the group `oarepo_workflows.state_changed_notifiers`.
        """
        group_name = "oarepo_workflows.state_changed_notifiers"
        return [
            x.load() for x in importlib_metadata.entry_points().select(group=group_name)
        ]

    @cached_property
    def workflow_changed_notifiers(self) -> list[WorkflowChangeNotifier]:
        """Return a list of workflow changed notifiers.

        Workflow changed notifiers are callables that are called when a workflow of a record changes.
        They are registered as entry points in the group `oarepo_workflows.workflow_changed_notifiers`.
        """
        group_name = "oarepo_workflows.workflow_changed_notifiers"
        return [
            x.load() for x in importlib_metadata.entry_points().select(group=group_name)
        ]

    @unit_of_work()
    def set_state(
        self,
        identity: Identity,
        record: Record,
        new_state: str,
        *args: Any,
        uow: UnitOfWork,
        commit: bool = True,
        **kwargs: Any,
    ) -> None:
        """Set a new state on a record.

        :param identity:    identity of the user who initiated the state change
        :param record:      record whose state is being changed
        :param new_state:   new state to set
        :param args:        additional arguments
        :param uow:         unit of work
        :param commit:      whether to commit the change
        :param kwargs:      additional keyword arguments
        """
        previous_value = record.state
        record.state = new_state
        if commit:
            service = get_record_service_for_record(record)
            uow.register(RecordCommitOp(record, indexer=service.indexer))
        for state_changed_notifier in self.state_changed_notifiers:
            state_changed_notifier(
                identity, record, previous_value, new_state, *args, uow=uow, **kwargs
            )

    @unit_of_work()
    def set_workflow(
        self,
        identity: Identity,
        record: Record,
        new_workflow_id: str,
        *args: Any,
        uow: UnitOfWork,
        commit: bool = True,
        **kwargs: Any,
    ) -> None:
        """Set a new workflow on a record.

        :param identity:            identity of the user who initiated the workflow change
        :param record:              record whose workflow is being changed
        :param new_workflow_id:     new workflow to set
        :param args:                additional arguments
        :param uow:                 unit of work
        :param commit:              whether to commit the change
        :param kwargs:              additional keyword arguments
        """
        if new_workflow_id not in current_oarepo_workflows.record_workflows:
            raise InvalidWorkflowError(
                f"Workflow {new_workflow_id} does not exist in the configuration.",
                record=record,
            )
        parent = record.parent  # noqa for typing: we do not have a better type for record with parent
        previous_value = parent.workflow
        parent.workflow = new_workflow_id
        if commit:
            service = get_record_service_for_record(record)
            uow.register(
                ParentRecordCommitOp(parent, indexer_context=dict(service=service))
            )
        for workflow_changed_notifier in self.workflow_changed_notifiers:
            workflow_changed_notifier(
                identity,
                record,
                previous_value,
                new_workflow_id,
                *args,
                uow=uow,
                **kwargs,
            )

    # noinspection PyMethodMayBeStatic
    def get_workflow_from_record(self, record: Record | ParentRecord) -> str | None:
        """Get the workflow from a record.

        :param record:  record to get the workflow from. Can pass either a record or its parent.
        """
        parent: ParentRecord = record.parent if hasattr(record, "parent") else record

        if hasattr(parent, "workflow") and parent.workflow:
            return parent.workflow
        else:
            return None

    @property
    def record_workflows(self) -> dict[str, Workflow]:
        """Return a dictionary of available record workflows."""
        return self.app.config["WORKFLOWS"]

    @property
    def default_workflow_events(self) -> dict:
        """Return a dictionary of default workflow events.

        Default workflow events are those that can be added to any request.
        The dictionary is taken from the configuration key `DEFAULT_WORKFLOW_EVENTS`.
        """
        return self.app.config.get("DEFAULT_WORKFLOW_EVENTS", {})

    def get_workflow(self, record: Record | dict[str, Any]) -> Workflow:
        """Get the workflow for a record.

        :param record:  record to get the workflow for
        :raises MissingWorkflowError: if the workflow is not found
        :raises InvalidWorkflowError: if the workflow is invalid
        """
        if hasattr(record, "parent"):
            try:
                record_parent: WithWorkflow = record.parent  # noqa for typing: we do not have a better type for record with parent
            except AttributeError as e:
                raise MissingWorkflowError(
                    "Record does not have a parent attribute, is it a draft-enabled record?",
                    record=record,
                ) from e
            try:
                workflow_id = record_parent.workflow
            except AttributeError as e:
                raise MissingWorkflowError(
                    "Parent record does not have a workflow attribute.", record=record
                ) from e
        else:
            try:
                dict_parent: dict[str, Any] = record["parent"]
            except KeyError as e:
                raise MissingWorkflowError(
                    "Record does not have a parent attribute.", record=record
                ) from e
            try:
                workflow_id = dict_parent["workflow"]
            except KeyError as e:
                raise MissingWorkflowError(
                    "Parent record does not have a workflow attribute.", record=record
                ) from e

        try:
            return self.record_workflows[workflow_id]
        except KeyError as e:
            raise InvalidWorkflowError(
                f"Workflow {workflow_id} doesn't exist in the configuration.",
                record=record,
            ) from e

    def init_app(self, app: Flask) -> None:
        """Flask application initialization."""
        # noinspection PyAttributeOutsideInit
        self.app = app
        app.extensions["oarepo-workflows"] = self


def finalize_app(app: Flask) -> None:
    """Finalize the application.

    This function registers the auto-approve service in the records resources registry.
    It is called from invenio_base.api_finalize_app entry point.

    :param app: Flask application
    """
    records_resources = app.extensions["invenio-records-resources"]

    ext: OARepoWorkflows = app.extensions["oarepo-workflows"]

    records_resources.registry.register(
        ext.auto_approve_service,
        service_id=ext.auto_approve_service.config.service_id,
    )

    records_resources.registry.register(
        ext.multiple_recipients_service,
        service_id=ext.multiple_recipients_service.config.service_id,
    )

    if app.config["OAREPO_WORKFLOWS_SET_REQUEST_PERMISSIONS"]:
        patch_request_permissions(app)


def patch_request_permissions(app: Flask) -> None:
    """Replace invenio request permissions.

    If permissions for requests are the plain invenio permissions,
    replace those with workflow-based ones. If user set their own
    permissions, keep those intact.
    """
    # patch invenio requests
    from invenio_requests.services.permissions import (
        PermissionPolicy as OriginalPermissionPolicy,
    )

    with app.app_context():
        from invenio_requests.proxies import current_requests_service

        from oarepo_workflows.requests.permissions import (
            CreatorsFromWorkflowRequestsPermissionPolicy,
        )

        current_permission_policy = app.config.get("REQUESTS_PERMISSION_POLICY")
        if current_permission_policy is OriginalPermissionPolicy:
            app.config["REQUESTS_PERMISSION_POLICY"] = (
                CreatorsFromWorkflowRequestsPermissionPolicy
            )
            assert (
                current_requests_service.config.permission_policy_cls
                is CreatorsFromWorkflowRequestsPermissionPolicy
            )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Model mixins."""

from __future__ import annotations

from invenio_db import db
from sqlalchemy import String


class RecordWorkflowParentModelMixin:
    """Mixin for a ParentRecord database model that adds a workflow field."""

    workflow = db.Column(String)
    """Workflow identifier."""



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""State system field."""

from __future__ import annotations

from datetime import UTC, datetime
from typing import Any, Optional, Protocol, Self, cast, overload

from invenio_records.systemfields.base import SystemField
from oarepo_runtime.records.systemfields import MappingSystemFieldMixin


class WithState(Protocol):
    """A protocol for a record containing a state field.

    Later on, if typing.Intersection is implemented,
    one could use it to have the record correctly typed as
    record: Intersection[WithState, Record]
    """

    state: str
    """State of the record."""

    state_timestamp: datetime
    """Timestamp of the last state change."""


class RecordStateField(MappingSystemFieldMixin, SystemField):
    """State system field."""

    def __init__(self, key: str = "state", initial: str = "draft") -> None:
        """Initialize the state field."""
        self._initial = initial
        super().__init__(key=key)

    def post_create(self, record: WithState) -> None:
        """Set the initial state when record is created."""
        self.set_dictkey(record, self._initial)

    def post_init(
        self, record: WithState, data: dict, model: Optional[Any] = None, **kwargs: Any
    ) -> None:
        """Set the initial state when record is created."""
        if not record.state:
            self.set_dictkey(record, self._initial)

    @overload
    def __get__(self, record: None, owner: type | None = None) -> Self: ...

    @overload
    def __get__(self, record: WithState, owner: type | None = None) -> str: ...

    def __get__(
        self, record: WithState | None, owner: type | None = None
    ) -> str | Self:
        """Get the persistent identifier."""
        if record is None:
            return self
        return self.get_dictkey(record)

    def __set__(self, record: WithState, value: str) -> None:
        """Directly set the state of the record."""
        if self.get_dictkey(record) != value:
            self.set_dictkey(record, value)
            cast(dict, record)["state_timestamp"] = datetime.now(tz=UTC).isoformat()

    @property
    def mapping(self) -> dict[str, dict[str, str]]:
        """Return the opensearch mapping for the state field."""
        return {
            self.attr_name: {"type": "keyword"},
        }


class RecordStateTimestampField(MappingSystemFieldMixin, SystemField):
    """State system field."""

    def __init__(self, key: str = "state_timestamp") -> None:
        """Initialize the state field."""
        super().__init__(key=key)

    def post_create(self, record: WithState) -> None:
        """Set the initial state when record is created."""
        self.set_dictkey(record, datetime.now(tz=UTC).isoformat())

    def post_init(
        self, record: WithState, data: dict, model: Optional[Any] = None, **kwargs: Any
    ) -> None:
        """Set the initial state when record is created."""
        if not record.state_timestamp:
            self.set_dictkey(record, datetime.now(tz=UTC).isoformat())

    @overload
    def __get__(self, record: None, owner: type | None = None) -> Self: ...

    @overload
    def __get__(self, record: WithState, owner: type | None = None) -> str: ...

    def __get__(
        self, record: WithState | None, owner: type | None = None
    ) -> str | Self:
        """Get the persistent identifier."""
        if record is None:
            return self
        return self.get_dictkey(record)

    @property
    def mapping(self) -> dict[str, dict[str, str]]:
        """Return the opensearch mapping for the state field."""
        # not needed as oarepo-model-builder-workflows already generated this field into the mapping
        return {
            self.attr_name: {
                "type": "date",
                "format": "strict_date_time||strict_date_time_no_millis||"
                "basic_date_time||basic_date_time_no_millis||"
                "basic_date||strict_date||"
                "strict_date_hour_minute_second||strict_date_hour_minute_second_fraction",
            },
        }



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Service for reading multiple-recipients entities."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, override

import marshmallow as ma
from oarepo_runtime.services.entity.config import EntityServiceConfig
from oarepo_runtime.services.entity.service import EntityService

if TYPE_CHECKING:
    from collections.abc import Iterable

    from flask_principal import Identity
    from invenio_records_resources.services.records.results import (
        RecordItem,
        RecordList,
    )


class MultipleEntitiesEntitySchema(ma.Schema):
    """Schema for multiple recipients entities."""

    entities = ma.fields.List(
        ma.fields.Dict(keys=ma.fields.String(), values=ma.fields.String())
    )


class MultipleEntitiesEntityServiceConfig(EntityServiceConfig):
    """Configuration for auto-approve entity service."""

    service_id = "multiple"
    schema = MultipleEntitiesEntitySchema


class MultipleEntitiesEntityService(EntityService):
    """Service for reading auto-approve entities."""

    @override
    def read(self, identity: Identity, id_: str, **kwargs: Any) -> RecordItem:
        return self.result_item(
            self,
            identity,
            record={"entities": json.loads(id_)},
            links_tpl=self.links_item_tpl,
        )

    @override
    def read_many(
        self,
        identity: Identity,
        ids: Iterable[str],
        fields: list[str] | None = None,
        **kwargs: Any,
    ) -> RecordList:
        results = []
        if ids:
            results = [{"entities": json.loads(id_)} for id_ in ids]
        return self.result_list(
            self,
            identity,
            results=results,
            links_item_tpl=self.links_item_tpl,
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Request workflow policy definition."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .requests import (
    WorkflowRequest,
)

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.records.api import Record


class WorkflowRequestPolicy:
    """Base class for workflow request policies.

    Inherit from this class
    and add properties to define specific requests for a workflow.

    The name of the property is the request_type name and the value must be
    an instance of WorkflowRequest.

    Example:
        class MyWorkflowRequests(WorkflowRequestPolicy):
            delete_request = WorkflowRequest(
                requesters = [
                    IfInState("published", RecordOwner())
                ],
                recipients = [CommunityRole("curator")],
                transitions: WorkflowTransitions(
                    submitted = 'considered_for_deletion',
                    approved = 'deleted',
                    rejected = 'published'
                )
            )

    """

    def __init__(self):
        """Initialize the request policy."""
        for rt_code, rt in self.items():
            rt._request_type = rt_code

    def __getitem__(self, request_type_id: str) -> WorkflowRequest:
        """Get the workflow request type by its id."""
        try:
            return getattr(self, request_type_id)
        except AttributeError:
            raise KeyError(
                f"Request type {request_type_id} not defined in {self.__class__.__name__}"
            ) from None

    def items(self) -> list[tuple[str, WorkflowRequest]]:
        """Return the list of request types and their instances.

        This call mimics mapping items() method.
        """
        ret = []
        parent_attrs = set(dir(WorkflowRequestPolicy))
        for attr in dir(self.__class__):
            if parent_attrs and attr in parent_attrs:
                continue
            if attr.startswith("_"):
                continue
            possible_request = getattr(self, attr, None)
            if isinstance(possible_request, WorkflowRequest):
                ret.append((attr, possible_request))
        return ret

    def applicable_workflow_requests(
        self, identity: Identity, *, record: Record, **context: Any
    ) -> list[tuple[str, WorkflowRequest]]:
        """Return a list of applicable requests for the identity and context.

        :param identity: Identity of the requester.
        :param context: Context of the request that is passed to the requester generators.
        :return: List of tuples (request_type_id, request) that are applicable for the identity and context.
        """
        ret = []

        for name, request in self.items():
            if request.is_applicable(identity, record=record, **context):
                ret.append((name, request))
        return ret



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Base permission policy that overwrites invenio-requests."""

from __future__ import annotations

from invenio_records_permissions.generators import AnyUser, SystemProcess
from invenio_requests.customizations.event_types import CommentEventType, LogEventType
from invenio_requests.services.generators import Creator, Receiver
from invenio_requests.services.permissions import (
    PermissionPolicy as InvenioRequestsPermissionPolicy,
)

from oarepo_workflows.requests.generators.conditionals import IfEventType
from oarepo_workflows.requests.generators.workflow_based import (
    EventCreatorsFromWorkflow,
    RequestCreatorsFromWorkflow,
)


class CreatorsFromWorkflowRequestsPermissionPolicy(InvenioRequestsPermissionPolicy):
    """Permissions for requests based on workflows.

    This permission adds a special generator RequestCreatorsFromWorkflow() to the default permissions.
    This generator takes a topic, gets the workflow from the topic and returns the generator for
    creators defined on the WorkflowRequest.
    """

    can_create = [
        SystemProcess(),
        RequestCreatorsFromWorkflow(),
    ]

    can_create_comment = [
        SystemProcess(),
        IfEventType(
            [LogEventType.type_id, CommentEventType.type_id], [Creator(), Receiver()]
        ),
        EventCreatorsFromWorkflow(),
    ]

    # any user can search for requests, but non-authenticated will not get a hit
    # this is necessary to not have errors on a secret link page (edit/preview form)
    # where the user is not authenticated and search for available requests is performed
    can_search = [AnyUser()]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Record policy for workflows."""

from __future__ import annotations

from functools import reduce

from invenio_records_permissions import RecordPermissionPolicy
from invenio_records_permissions.generators import (
    AnyUser,
    SystemProcess,
)
from invenio_search.engine import dsl

from ...proxies import current_oarepo_workflows
from .generators import FromRecordWorkflow


class WorkflowRecordPermissionPolicy(RecordPermissionPolicy):
    """Permission policy to be used in permission presets directly on RecordServiceConfig.permission_policy_cls.

    Do not use this class in Workflow constructor.
    """

    can_commit_files = [FromRecordWorkflow("commit_files")]
    can_create = [FromRecordWorkflow("create")]
    can_create_files = [FromRecordWorkflow("create_files")]
    can_delete = [FromRecordWorkflow("delete")]
    can_delete_draft = [FromRecordWorkflow("delete_draft")]
    can_delete_files = [FromRecordWorkflow("delete_files")]
    can_draft_create_files = [FromRecordWorkflow("draft_create_files")]
    can_edit = [FromRecordWorkflow("edit")]
    can_get_content_files = [FromRecordWorkflow("get_content_files")]
    can_list_files = [FromRecordWorkflow("list_files")]
    can_new_version = [FromRecordWorkflow("new_version")]
    can_publish = [FromRecordWorkflow("publish")]
    can_read = [FromRecordWorkflow("read")]
    can_read_all_records = [FromRecordWorkflow("read_all_records")]
    can_read_deleted = [FromRecordWorkflow("read_deleted")]
    can_read_draft = [FromRecordWorkflow("read_draft")]
    can_read_files = [FromRecordWorkflow("read_files")]
    can_search = [SystemProcess(), AnyUser()]
    can_search_drafts = [SystemProcess(), AnyUser()]
    can_search_versions = [SystemProcess(), AnyUser()]
    can_search_all_records = [SystemProcess(), AnyUser()]
    can_set_content_files = [FromRecordWorkflow("set_content_files")]
    can_update = [FromRecordWorkflow("update")]
    can_update_draft = [FromRecordWorkflow("update_draft")]
    can_update_files = [FromRecordWorkflow("update_files")]

    # extra from rdm - some of these might we might ignore, but adding them here
    # will not break anything
    can_add_community = [FromRecordWorkflow("add_community")]
    can_bulk_add = [FromRecordWorkflow("bulk_add")]
    can_draft_commit_files = [FromRecordWorkflow("draft_commit_files")]
    can_draft_delete_files = [FromRecordWorkflow("draft_delete_files")]
    can_draft_get_content_files = [FromRecordWorkflow("draft_get_content_files")]
    can_draft_media_commit_files = [FromRecordWorkflow("draft_media_commit_files")]
    can_draft_media_create_files = [FromRecordWorkflow("draft_media_create_files")]
    can_draft_media_delete_files = [FromRecordWorkflow("draft_media_delete_files")]
    can_draft_media_get_content_files = [
        FromRecordWorkflow("draft_media_get_content_files")
    ]
    can_draft_media_read_files = [FromRecordWorkflow("draft_media_read_files")]
    can_draft_media_set_content_files = [
        FromRecordWorkflow("draft_media_set_content_files")
    ]
    can_draft_media_update_files = [FromRecordWorkflow("draft_media_update_files")]
    can_draft_read_files = [FromRecordWorkflow("draft_read_files")]
    can_draft_set_content_files = [FromRecordWorkflow("draft_set_content_files")]
    can_draft_update_files = [FromRecordWorkflow("draft_update_files")]
    can_lift_embargo = [FromRecordWorkflow("lift_embargo")]
    can_manage = [FromRecordWorkflow("manage")]
    can_manage_files = [FromRecordWorkflow("manage_files")]
    can_manage_internal = [FromRecordWorkflow("manage_internal")]
    can_manage_quota = [FromRecordWorkflow("manage_quota")]
    can_manage_record_access = [FromRecordWorkflow("manage_record_access")]
    can_media_commit_files = [FromRecordWorkflow("media_commit_files")]
    can_media_create_files = [FromRecordWorkflow("media_create_files")]
    can_media_delete_files = [FromRecordWorkflow("media_delete_files")]
    can_media_get_content_files = [FromRecordWorkflow("media_get_content_files")]
    can_media_read_deleted_files = [FromRecordWorkflow("media_read_deleted_files")]
    can_media_read_files = [FromRecordWorkflow("media_read_files")]
    can_media_set_content_files = [FromRecordWorkflow("media_set_content_files")]
    can_media_update_files = [FromRecordWorkflow("media_update_files")]
    can_moderate = [FromRecordWorkflow("moderate")]
    can_pid_create = [FromRecordWorkflow("pid_create")]
    can_pid_delete = [FromRecordWorkflow("pid_delete")]
    can_pid_discard = [FromRecordWorkflow("pid_discard")]
    can_pid_manage = [FromRecordWorkflow("pid_manage")]
    can_pid_register = [FromRecordWorkflow("pid_register")]
    can_pid_update = [FromRecordWorkflow("pid_update")]
    can_preview = [FromRecordWorkflow("preview")]
    can_purge = [FromRecordWorkflow("purge")]
    can_query_stats = [FromRecordWorkflow("query_stats")]
    can_read_deleted_files = [FromRecordWorkflow("read_deleted_files")]
    can_remove_community = [FromRecordWorkflow("remove_community")]
    can_remove_record = [FromRecordWorkflow("remove_record")]
    can_review = [FromRecordWorkflow("review")]
    can_view = [FromRecordWorkflow("view")]

    @property
    def query_filters(self) -> list[dict]:
        """Return query filters from the delegated workflow permissions."""
        if self.action not in (
            "read",
            "read_draft",
            "read_deleted",
            "read_all_records",
        ):
            return super().query_filters
        workflows = current_oarepo_workflows.record_workflows
        queries = []
        for workflow_id, workflow in workflows.items():
            q_in_workflow = dsl.Q("term", **{"parent.workflow": workflow_id})
            workflow_filters = workflow.permissions(
                self.action, **self.over
            ).query_filters
            if not workflow_filters:
                workflow_filters = [dsl.Q("match_none")]
            query = reduce(lambda f1, f2: f1 | f2, workflow_filters) & q_in_workflow
            queries.append(query)
        return [q for q in queries if q]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Generator that combines multiple generators together with an 'or' operation."""

from __future__ import annotations

import dataclasses
import json
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional, override

from invenio_records_permissions.generators import Generator

from .recipient_generator import RecipientGeneratorMixin

if TYPE_CHECKING:
    from flask_principal import Need
    from invenio_records_resources.records.api import Record
    from invenio_requests.customizations.request_types import RequestType


@dataclasses.dataclass
class MultipleEntitiesGenerator(RecipientGeneratorMixin, Generator):
    """A generator that combines multiple generators with 'or' operation."""

    generators: list[Generator] | tuple[Generator]
    """List of generators to be combined."""

    @override
    def needs(self, **context: Any) -> set[Need]:
        """Generate a set of needs from generators that a person needs to have.

        :param context: Context.
        :return: Set of needs.
        """
        return {
            need for generator in self.generators for need in generator.needs(**context)
        }

    @override
    def excludes(self, **context: Any) -> set[Need]:
        """Generate a set of needs that person must not have.

        :param context: Context.
        :return: Set of needs.
        """
        return {
            exclude
            for generator in self.generators
            for exclude in generator.excludes(**context)
        }

    @override
    def query_filter(self, **context: Any) -> list[dict]:
        """Generate a list of opensearch query filters.

         These filters are used to filter objects. These objects are governed by a policy
         containing this generator.

        :param context: Context.
        """
        ret: list[dict] = []
        for generator in self.generators:
            query_filter = generator.query_filter(**context)
            if query_filter:
                if isinstance(query_filter, Iterable):
                    ret.extend(query_filter)
                else:
                    ret.append(query_filter)
        return ret

    @override
    def reference_receivers(
        self,
        record: Optional[Record] = None,
        request_type: Optional[RequestType] = None,
        **context: Any,
    ) -> list[dict[str, str]]:  # pragma: no cover
        """Return the reference receiver(s) of the request.

        This call requires the context to contain at least "record" and "request_type"

        Must return a list of dictionary serialization of the receivers.

        Might return empty list or None to indicate that the generator does not
        provide any receivers.
        """
        references = []

        for generator in self.generators:
            if not isinstance(generator, RecipientGeneratorMixin):
                raise ValueError(
                    f"Generator {generator} is not a recipient generator and can not be used in "
                    f"MultipleGeneratorsGenerator."
                )

            reference = generator.reference_receivers(
                record=record, request_type=request_type, **context
            )
            if reference:
                references.extend(reference)
        if not references:
            return []
        if len(references) == 1:
            return references

        return [{"multiple": json.dumps(references)}]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Record layer, system fields."""

from __future__ import annotations

from .state import RecordStateField, RecordStateTimestampField
from .workflow import WorkflowField

__all__ = (
    "RecordStateField",
    "WorkflowField",
    "RecordStateTimestampField",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Permission generators usable in workflow configurations."""

from __future__ import annotations

import operator
from functools import reduce
from itertools import chain
from typing import TYPE_CHECKING, Any, Optional

from invenio_records_permissions.generators import ConditionalGenerator, Generator
from invenio_search.engine import dsl

from oarepo_workflows.errors import InvalidWorkflowError, MissingWorkflowError
from oarepo_workflows.proxies import current_oarepo_workflows

if TYPE_CHECKING:
    from collections.abc import Iterable

    from flask_principal import Need
    from invenio_records_permissions import RecordPermissionPolicy
    from invenio_records_resources.records import Record


# invenio_records_permissions.generators.ConditionalGenerator._make_query
def _make_query(generators: Iterable[Generator], **context: Any) -> dict | None:
    queries = [g.query_filter(**context) for g in generators]
    queries = [q for q in queries if q]
    return reduce(operator.or_, queries) if queries else dsl.Q("match_none")


class FromRecordWorkflow(Generator):
    """Permission delegating check to workflow.

    The implementation of the permission gets the workflow id from the passed context
    (record or data) and then looks up the workflow definition in the configuration.

    The workflow definition must contain a permissions policy that is then used to
    determine the permissions for the action.
    """

    _action: str

    def __init__(self, action: str) -> None:
        """Initialize the permission."""
        # might not be needed in subclasses
        super().__init__()
        self._action = action

    # noinspection PyMethodMayBeStatic
    def _get_workflow_id(self, record: Optional[Record] = None, **context: Any) -> str:
        """Get the workflow id from the context.

        If the record is passed, the workflow is determined from the record.
        If the record is not passed, the workflow is determined from the input data.

        If the workflow is not found, an error is raised.

        :param record: Record to get the workflow from.
        :param context: Context to get the workflow from.
        :return: Workflow id.
        :raises MissingWorkflowError: If the workflow is not found on the record/data.
        """
        if record:
            workflow_id = current_oarepo_workflows.get_workflow_from_record(record)
            if not workflow_id:
                raise MissingWorkflowError(
                    "Workflow not defined on record.", record=record
                )
        else:
            data = context.get("data", {})
            workflow_id = data.get("parent", {}).get("workflow", {})
            if not workflow_id:
                raise MissingWorkflowError(
                    "Workflow not defined in input.", record=data
                )
        return workflow_id

    def _get_permissions_from_workflow(
        self,
        action_name: str,
        record: Optional[Record] = None,
        **context: Any,
    ) -> RecordPermissionPolicy:
        """Get the permissions policy from the workflow.

        At first the workflow id is determined from the context.
        Then the permissions policy is determined from the workflow configuration,
        is instantiated with the action name and the context and the permissions
        for the action are returned.
        """
        workflow_id = self._get_workflow_id(record, **context)
        if workflow_id not in current_oarepo_workflows.record_workflows:
            raise InvalidWorkflowError(
                f"Workflow {workflow_id} does not exist in the configuration.",
                record=record or context.get("data", {}),
            )
        policy = current_oarepo_workflows.record_workflows[workflow_id].permissions
        ret = policy(action_name, record=record, **context)
        # debug point
        return ret

    def needs(self, record: Optional[Record] = None, **context: Any) -> list[Need]:
        """Return needs that are generated by the workflow permission."""
        return self._get_permissions_from_workflow(
            self._action, record, **context
        ).needs

    def excludes(self, **context: Any) -> list[Need]:
        """Return excludes that are generated by the workflow permission."""
        return self._get_permissions_from_workflow(self._action, **context).excludes

    def query_filter(self, record: Optional[Record] = None, **context: Any) -> dict:
        """Return query filters that are generated by the workflow permission.

        Note: this implementation in fact will be called from WorkflowRecordPermissionPolicy.query_filters
        for each registered workflow type. The query_filters are then combined into a single query.
        """
        return self._get_permissions_from_workflow(
            self._action, record, **context
        ).query_filters


class WorkflowPermission(FromRecordWorkflow):
    """Deprecated alias for FromRecordWorkflow."""

    def __init__(self, action: str) -> None:
        """Initialize the generator."""
        import warnings

        warnings.warn(
            "WorkflowPermission is deprecated. Use FromRecordWorkflow instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(action)


class IfInState(ConditionalGenerator):
    """Generator that checks if the record is in a specific state.

    If it is in the state, the then_ generators are used, otherwise the else_ generators are used.

    Example:
        .. code-block:: python

            can_edit = [ IfInState("draft", [RecordOwners()]) ]

    """

    def __init__(
        self,
        state: str | list[str] | tuple[str, ...],
        then_: list[Generator] | tuple[Generator] | Generator | None = None,
        else_: list[Generator] | tuple[Generator] | Generator | None = None,
    ) -> None:
        """Initialize the generator."""
        if isinstance(then_, Generator):
            then_ = [then_]
        if isinstance(else_, Generator):
            else_ = [else_]
        super().__init__(then_ or [], else_=else_ or [])
        if isinstance(state, str):
            state = [state]
        if not isinstance(state, (list, tuple)):
            raise TypeError(
                f"State must be a string, list or tuple. Got {type(state)}."
            )
        self.state = state

    def _condition(self, record: Record, **context: Any) -> bool:
        """Check if the record is in the state."""
        try:
            return record.state in self.state  # noqa as AttributeError is caught
        except AttributeError:
            return False

    def query_filter(self, **context: Any) -> dsl.Q:
        """Apply then or else filter."""
        field = "state"

        q_instate = dsl.Q("terms", **{field: self.state})
        if self.then_:
            then_query = self._make_query(self.then_, **context)
        else:
            then_query = dsl.Q("match_none")

        if self.else_:
            else_query = self._make_query(self.else_, **context)
        else:
            else_query = dsl.Q("match_none")

        return (q_instate & then_query) | (~q_instate & else_query)

    def __repr__(self) -> str:
        """Representation of the generator."""
        return (
            f"IfInState({self.state}, then={repr(self.then_)}, else={repr(self.else_)})"
        )

    def __str__(self) -> str:
        """String representation of the generator."""
        return repr(self)


class SameAs(Generator):
    """Generator that delegates the permissions to another action.

    Example:
        .. code-block:: python
            class Perms:
                can_create_files = [SameAs("edit_files")]
                can_edit_files = [RecordOwners()]

    would mean that the permissions for creating files are the same as for editing files.
    This works even if you inherit from the class and override the can_edit_files.

    """

    def __init__(self, permission_name: str) -> None:
        """Initialize the generator.

        :param permission_name: Name of the permission to delegate to. In most cases,
        it will look like "can_<action>". A property with this name must exist on the policy
        and its value must be a list of generators.
        """
        self.delegated_permission_name = permission_name

    # noinspection PyUnusedLocal
    def _generators(
        self, *, policy: RecordPermissionPolicy, **context: Any
    ) -> list[Generator]:
        """Get the generators from the policy."""
        return getattr(policy, self.delegated_permission_name)

    def needs(
        self, *, policy: RecordPermissionPolicy | None = None, **context: Any
    ) -> list[Need]:
        """Get the needs from the policy."""
        if policy is None:
            raise ValueError(
                f"SameAs: Policy must be passed to the generator. Got the following context: {context.keys()}"
            )
        needs = [
            generator.needs(policy=policy, **context)
            for generator in self._generators(policy=policy, **context)
        ]
        return list(chain.from_iterable(needs))

    def excludes(
        self, *, policy: RecordPermissionPolicy | None = None, **context: Any
    ) -> list[Need]:
        """Get the excludes from the policy."""
        if policy is None:
            raise ValueError(
                f"SameAs: Policy must be passed to the generator. Got the following context: {context.keys()}"
            )
        excludes = [
            generator.excludes(policy=policy, **context)
            for generator in self._generators(policy=policy, **context)
        ]
        return list(chain.from_iterable(excludes))

    def query_filter(
        self, *, policy: RecordPermissionPolicy, **context: Any
    ) -> dict | None:
        """Search filters."""
        return _make_query(self._generators(policy=policy, **context), **context)

    def __repr__(self) -> str:
        """Representation of the generator."""
        return f"SameAs({self.delegated_permission_name})"

    def __str__(self) -> str:
        """String representation of the generator."""
        return repr(self)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Service components for supporting workflows on Invenio records."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from invenio_records_resources.services.records.components.base import ServiceComponent

from oarepo_workflows.errors import MissingWorkflowError
from oarepo_workflows.proxies import current_oarepo_workflows

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.records import Record


class WorkflowSetupComponent(ServiceComponent):
    pass


class WorkflowComponent(ServiceComponent):
    """Workflow component.

    This component is responsible for checking if the workflow is defined in the input data
    when record is created. If it is not present, it raises an error.
    """

    depends_on = [WorkflowSetupComponent]
    affects = "*"
    # put it before metadata component to make sure that the workflow
    # is set at the beginning of the record creation process. This makes sure
    # that for example file checks are done with the correct workflow set.

    def create(
        self,
        identity: Identity,
        data: Optional[dict[str, Any]] = None,
        record: Optional[Record] = None,
        **kwargs: Any,
    ) -> None:
        """Implement record creation checks and set the workflow on the created record."""
        if not data:
            # sanity check, should be handled by policy before the component is called
            raise MissingWorkflowError(
                "Workflow not defined in input. As this should be handled by a policy, "
                "make sure you are using workflow-enabled policy.",
                record=data,
            )  # pragma: no cover
        try:
            workflow_id = data["parent"]["workflow"]
        except KeyError as e:  # pragma: no cover
            # sanity check, should be handled by policy before the component is called
            raise MissingWorkflowError(  # pragma: no cover
                "Workflow not defined in input. As this should be handled by a policy, "
                "make sure you are using workflow-enabled policy.",
                record=data,
            ) from e

        current_oarepo_workflows.set_workflow(
            identity, record, workflow_id, uow=self.uow, **kwargs
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Default configuration for oarepo-workflows."""

from __future__ import annotations

from typing import TYPE_CHECKING

from oarepo_workflows.services.permissions.record_permission_policy import (
    WorkflowRecordPermissionPolicy,
)

if TYPE_CHECKING:
    from oarepo_workflows import Workflow

OAREPO_PERMISSIONS_PRESETS = {
    "workflow": WorkflowRecordPermissionPolicy,
}
"""Permissions presets for oarepo-workflows."""

WORKFLOWS: dict[str, Workflow] = {}
"""Configuration of workflows, must be provided by the user inside, for example, invenio.cfg."""

OAREPO_WORKFLOWS_SET_REQUEST_PERMISSIONS = True

WORKFLOWS_ALLOWED_REQUEST_RECEIVERS = ["multiple"]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Support any workflows on Invenio record."""

from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any, Protocol

from . import WorkflowRecordPermissionPolicy
from .requests import WorkflowRequestPolicy
from .services.permissions import DefaultWorkflowPermissions

if TYPE_CHECKING:
    from flask_babel import LazyString
    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_records_resources.services.uow import UnitOfWork


@dataclasses.dataclass
class Workflow:
    """A workflow definition."""

    label: str | LazyString
    """A human-readable label for the workflow."""

    permission_policy_cls: type[DefaultWorkflowPermissions]
    """A permission policy class that handles permissions on records that are governed by this workflow."""

    request_policy_cls: type[WorkflowRequestPolicy] = WorkflowRequestPolicy
    """A request policy class that defines which requests can be applied to records governed by this workflow."""

    def permissions(self, action: str, **over: Any) -> DefaultWorkflowPermissions:
        """Return permission policy for this workflow applicable to the given action.

        :param  action:      action for which permission is asked
        """
        return self.permission_policy_cls(action, **over)

    def requests(self) -> WorkflowRequestPolicy:
        """Return instance of request policy for this workflow."""
        return self.request_policy_cls()

    def __post_init__(self) -> None:
        """Check that the classes are subclasses of the expected classes.

        This is just a sanity check to raise an error as soon as possible.
        """
        assert not issubclass(
            self.permission_policy_cls, WorkflowRecordPermissionPolicy
        )
        assert issubclass(self.permission_policy_cls, DefaultWorkflowPermissions)
        assert issubclass(self.request_policy_cls, WorkflowRequestPolicy)


class StateChangedNotifier(Protocol):
    """A protocol for a state change notifier.

    State changed notifier is a callable that is called when a state of a record changes,
    for example as a result of a workflow transition.
    """

    def __call__(
        self,
        identity: Identity,
        record: Record,
        previous_state: str,
        new_state: str,
        *args: Any,
        uow: UnitOfWork,
        **kwargs: Any,
    ):
        """Notify about a state change.

        :param identity:        identity of the user who initiated the state change
        :param record:          record whose state changed
        :param previous_state:  previous state of the record
        :param new_state:       new state of the record
        :param args:            additional arguments
        :param uow:             unit of work
        :param kwargs:          additional keyword arguments
        """
        ...


class WorkflowChangeNotifier(Protocol):
    """A protocol for a workflow change notifier.

    Workflow changed notifier is a callable that is called
    when a workflow of a record changes.
    """

    def __call__(
        self,
        identity: Identity,
        record: Record,
        previous_workflow_id: str,
        new_workflow_id: str,
        *args: Any,
        uow: UnitOfWork,
        **kwargs: Any,
    ):
        """Notify about a workflow change.

        :param identity:                identity of the user who initiated the workflow change
        :param record:                  record whose workflow changed
        :param previous_workflow_id:    previous workflow of the record
        :param new_workflow_id:         new workflow of the record
        :param args:                    additional arguments
        :param uow:                     unit of work
        :param kwargs:                  additional keyword arguments
        """
        ...



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-workflows is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Workflow requests."""

from __future__ import annotations

import dataclasses
from functools import cached_property
from logging import getLogger
from typing import TYPE_CHECKING, Any, Optional

from invenio_requests.proxies import (
    current_request_type_registry,
    current_requests_service,
)

from oarepo_workflows.errors import InvalidConfigurationError
from oarepo_workflows.proxies import current_oarepo_workflows
from oarepo_workflows.requests.generators.multiple_entities import (
    MultipleEntitiesGenerator,
)

if TYPE_CHECKING:
    from datetime import timedelta

    from flask_principal import Identity
    from invenio_records_permissions.generators import Generator
    from invenio_records_resources.records.api import Record
    from invenio_requests.customizations.request_types import RequestType

    from oarepo_workflows.requests.events import WorkflowEvent

log = getLogger(__name__)


@dataclasses.dataclass
class WorkflowRequest:
    """Workflow request definition.

    The request is defined by the requesters and recipients.
    The requesters are the generators that define who can submit the request. The recipients
    are the generators that define who can approve the request.
    """

    requesters: list[Generator] | tuple[Generator]
    """Generators that define who can submit the request."""

    recipients: list[Generator] | tuple[Generator]
    """Generators that define who can approve the request."""

    events: dict[str, WorkflowEvent] = dataclasses.field(default_factory=lambda: {})
    """Events that can be submitted with the request."""

    transitions: Optional[WorkflowTransitions] = dataclasses.field(
        default_factory=lambda: WorkflowTransitions()
    )
    """Transitions applied to the state of the topic of the request."""

    escalations: Optional[list[WorkflowRequestEscalation]] = None
    """Escalations applied to the request if not approved/declined in time."""

    _request_type: RequestType | None = dataclasses.field(default=None, init=False)

    @cached_property
    def requester_generator(self) -> Generator:
        """Return the requesters as a single requester generator."""
        return MultipleEntitiesGenerator(self.requesters)

    def recipient_entity_reference(self, **context: Any) -> dict | None:
        """Return the reference receiver of the workflow request with the given context.

        Note: invenio supports only one receiver, so the first one is returned at the moment.
        Later on, a composite receiver can be implemented.

        :param context: Context of the request.
        """
        return RecipientEntityReference(self, **context)

    def is_applicable(
        self, identity: Identity, *, record: Record, **context: Any
    ) -> bool:
        """Check if the request is applicable for the identity and context (which might include record, community, ...).

        :param identity: Identity of the requester.
        :param context: Context of the request that is passed to the requester generators.
        """
        try:
            if hasattr(self.request_type, "is_applicable_to"):
                # the is_applicable_to must contain a permission check, so do not need to do any check here ...
                return self.request_type.is_applicable_to(
                    identity, topic=record, **context
                )
            else:
                return current_requests_service.check_permission(
                    identity,
                    "create",
                    record=record,
                    request_type=self.request_type,
                    **context,
                )
        except InvalidConfigurationError:
            raise
        except Exception as e:
            log.exception("Error checking request applicability: %s", e)
            return False

    @cached_property
    def request_type(self) -> RequestType:
        """Return the request type for the workflow request."""
        if self._request_type is None:
            raise InvalidConfigurationError(
                f"Probably this WorkflowRequest ({self}) is not part of a Workflow. "
                "Please add it to a workflow or manually set the ._request_type attribute "
                "to a code of a registered request type."
            )
        try:
            return current_request_type_registry.lookup(self._request_type)
        except KeyError as e:
            raise InvalidConfigurationError(
                f"Request type {self._request_type} not found in the request type registry."
            ) from e

    @property
    def allowed_events(self) -> dict[str, WorkflowEvent]:
        """Return the allowed events for the workflow request."""
        return current_oarepo_workflows.default_workflow_events | self.events


@dataclasses.dataclass
class WorkflowTransitions:
    """Transitions for a workflow request.

    If the request is submitted and submitted is filled,
    the record (topic) of the request will be moved to state defined in submitted.
    If the request is approved, the record will be moved to state defined in approved.
    If the request is rejected, the record will be moved to state defined in rejected.
    """

    submitted: Optional[str] = None
    accepted: Optional[str] = None
    declined: Optional[str] = None
    cancelled: Optional[str] = None

    def __getitem__(self, transition_name: str):
        """Get the transition by name."""
        if transition_name not in ["submitted", "accepted", "declined", "cancelled"]:
            raise KeyError(
                f"Transition {transition_name} not defined in {self.__class__.__name__}"
            )
        return getattr(self, transition_name)


@dataclasses.dataclass
class WorkflowRequestEscalation:
    """Escalation of the request.

    If the request is not approved/declined/cancelled in time, it might be passed to another recipient
    (such as a supervisor, administrator, ...). The escalation is defined by the time after which the
    request is escalated and the recipients of the escalation.
    """

    after: timedelta
    recipients: list[Generator] | tuple[Generator]

    @property
    def escalation_id(self):
        """Return the escalation ID."""
        return str(self.after.total_seconds())

    def __repr__(self):
        """Representation of the WorkflowRequestEscalation."""
        return str(self)

    def __str__(self):
        """String representation of WorkflowRequestEscalation containing after time and recipients."""
        return f'{self.after=},{self.recipients=}'

# noinspection PyPep8Naming
def RecipientEntityReference(request_or_escalation: WorkflowRequest | WorkflowRequestEscalation,
                             **context: Any) -> dict | None:
    """Return the reference receiver of the workflow request or workflow request escalation with the given context.

    Note: invenio supports only one receiver, so the first one is returned at the moment.
    Later on, a composite receiver can be implemented.

    :param request_or_escalation: Workflow request or WorkflowRequestEscalation.
    :param context: Context of the request.

    :return: Reference receiver as a dictionary or None if no receiver has been resolved.

    Implementation note: intentionally looks like a class, later on might be converted into one extending from dict.
    """
    if not request_or_escalation.recipients:
        return None

    generator = MultipleEntitiesGenerator(request_or_escalation.recipients)
    receivers = generator.reference_receivers(**context)
    return receivers[0] if receivers else None
